@@ -75,9 +75,10 @@ import ogAST
 
 LOG = logging.getLogger(__name__)
 
-# reference to the ASN.1 Data view and to the process variables
+# reference to the ASN.1 Data view and to the visible variables (in scope)
 TYPES = None
 VARIABLES = {}
+LOCAL_VAR = {}
 # List of output signals and procedures
 OUT_SIGNALS = []
 PROCEDURES = []
@@ -86,21 +87,30 @@ INNER_PROCEDURES = []
 # lookup table for mapping SDL operands with the corresponding Ada ones
 OPERANDS = {'plus': '+', 'mul': '*', 'minus': '-', 'or': 'or',
         'and': 'and', 'xor': 'CHECKME', 'eq': '=', 'neq': '/=', 'gt': '>',
-        'ge': '>=', 'lt': '<', 'le': '<=', 'div': '/', 'mod': 'mod'}
+        'ge': '>=', 'lt': '<', 'le': '<=', 'div': '/', 'mod': 'mod',
+        'rem': 'mod'}
 
 
 def find_basic_type(a_type):
     ''' Return the ASN.1 basic type of aType '''
     basic_type = a_type
-    while basic_type['Kind'] == 'ReferenceType':
+    while basic_type.kind == 'ReferenceType':
         # Find type with proper case in the data view
         for typename in TYPES.viewkeys():
-            if typename.lower() == basic_type['ReferencedTypeName'].lower():
-                basic_type = TYPES[typename]['type']
+            if typename.lower() == basic_type.ReferencedTypeName.lower():
+                basic_type = TYPES[typename].type
                 break
     return basic_type
 
 
+def find_var(var):
+    ''' Return a variable from the scope, with proper case '''
+    for visible_var in VARIABLES.viewkeys():
+        if var.lower() == visible_var.lower():
+            return visible_var
+    return None
+
+
 def path_type(path):
     '''
         Return the type of a path construct
@@ -110,12 +120,12 @@ def path_type(path):
     '''
     # NOTE: all this code is duplicated from the find_type function
     # in ogParser - should be reused
-    if not path or not path[0].lower() in VARIABLES:
+    if not path or not find_var(path[0]):
         return None, None
     kind = ''
-    asn1_name = VARIABLES[path[0].lower()].replace('_', '-')
+    asn1_name = VARIABLES[find_var(path[0])].ReferencedTypeName
     # Get ASN.1 type of the first element
-    current = TYPES[asn1_name]['type']
+    current = TYPES[asn1_name].type
     if len(path) > 1:
         for elem in path[1:]:
             current = find_basic_type(current)
@@ -123,24 +133,24 @@ def path_type(path):
                 # Discard operator parameters: they do not change the type
                 continue
             # Sequence, Choice (case insensitive)
-            if current['Kind'] in ('SequenceType', 'ChoiceType'):
+            if current.kind in ('SequenceType', 'ChoiceType'):
                 elem_asn1 = elem.replace('_', '-').lower()
-                type_idx, = (c for c in current['Children']
+                type_idx, = (c for c in current.Children
                                     if c.lower() == elem_asn1)
-                current = current['Children'][type_idx]['type']
+                current = current.Children[type_idx].type
             # Sequence of
-            elif current['Kind'] == 'SequenceOfType':
+            elif current.kind == 'SequenceOfType':
                 current = current['type']
-            elif current['Kind'] == 'EnumeratedType':
+            elif current.kind.endswith('EnumeratedType'):
                 pass
             else:
                 raise TypeError('Expression ' + '.'.join(path) +
                                 ' does not resolve')
-        asn1_name = current['ReferencedTypeName'].replace('-', '_')
-    kind = find_basic_type(current).get('Kind').replace('-', '_')
+        asn1_name = current.ReferencedTypeName
+    kind = find_basic_type(current).kind.replace('-', '_')
     LOG.debug('[Ada Generator] Type of path ' + '!'.join(path) +
              ' is ' + asn1_name + ' (' + kind + ')')
-    return kind, asn1_name
+    return kind, asn1_name.replace('-', '_')
 
 
 def traceability(symbol):
@@ -158,7 +168,7 @@ def write_statement(param, newline):
     string = ''
     local = []
     basic_type = find_basic_type(param.exprType) or {}
-    type_kind = basic_type.get('Kind')
+    type_kind = basic_type.kind
     if type_kind.endswith('StringType'):
         if param.var.kind == 'stringLiteral':
             # Raw string
@@ -262,6 +272,8 @@ def output_statement(output):
                     param_type = out_sig.fpar[idx]['type']
                     param_direction = out_sig.fpar[idx]['direction']
 
+                typename = param_type.ReferencedTypeName.replace('-', '_')
+
                 p_code, p_id, p_local = decipher_expression(param)
                 code.extend(p_code)
                 local_decl.extend(p_local)
@@ -269,8 +281,7 @@ def output_statement(output):
                 if param_direction == 'in':
                     tmp_id = out['tmpVars'][idx]
                     local_decl.append('tmp{idx} : aliased asn1Scc{oType};'
-                                      .format(idx=tmp_id,
-                                          oType=param_type.replace('-', '_')))
+                                      .format(idx=tmp_id, oType=typename))
                     code.append('tmp{idx} := {p_id};'.format(
                         idx=tmp_id, p_id=p_id))
                     list_of_params.append("tmp{idx}'access".
@@ -297,7 +308,7 @@ def output_statement(output):
                     proc=proc.inputString,
                     params=', '.join(list_of_params)))
             else:
-                code.append('{proc};'.format(proc.inputString))
+                code.append('{};'.format(proc.inputString))
     return code, local_decl
 
 
@@ -350,10 +361,10 @@ def decipher_primary_id(primaryId):
     ada_string = ''
     stmts = []
     local_decl = []
-   
+
     # If first element is not a variable (can be a timer) do not add prefix
-    sep = 'l_' if primaryId[0].lower() in VARIABLES else ''
-    
+    sep = 'l_' if find_var(primaryId[0]) else ''
+
     sub_id = []
     for pr_id in primaryId:
         if type(pr_id) is not dict:
@@ -368,7 +379,7 @@ def decipher_primary_id(primaryId):
             sub_id.append(pr_id)
             if parent_kind == 'ChoiceType':
                 ada_string = ('asn1Scc{typename}_{p_id}_get({ada_string})'
-                        .format(typename=parent_typename.replace('-', '_'),
+                        .format(typename=parent_typename,
                                 p_id=pr_id, ada_string=ada_string))
             else:
                 ada_string += sep + pr_id
@@ -385,12 +396,12 @@ def decipher_primary_id(primaryId):
                 stmts.extend(r1_stmts)
                 stmts.extend(r2_stmts)
                 local_decl.extend(r1_local)
-                local_decl.extend(r2_local)               
-                _, parent_typename = path_type(sub_id) 
+                local_decl.extend(r2_local)
+                _, parent_typename = path_type(sub_id)
                 local_decl.append('tmp{idx} : aliased asn1Scc{parent_type};'
                           .format(idx=pr_id['tmpVar'],
-                              parent_type=parent_typename.replace('-', '_')))
-                  # XXX types with fixed length: substrings will not work
+                              parent_type=parent_typename))
+                # XXX types with fixed length: substrings will not work
                 if unicode.isnumeric(r1_string) and unicode.isnumeric(
                                                                 r2_string):
                     length = eval(r2_string + '-' + r1_string + '+1')
@@ -420,8 +431,8 @@ def decipher_primary_id(primaryId):
                     # Length of sequence of: take only the first parameter
                     exp, = pr_id['procParams']
                     exp_type = find_basic_type(exp.exprType)
-                    min_length = exp_type.get('Min')
-                    max_length = exp_type.get('Max')
+                    min_length = getattr(exp_type, 'Min', None)
+                    max_length = getattr(exp_type, 'Max', None)
                     if min_length is None or max_length is None:
                         error = '{} is not a SEQUENCE OF'.format(
                                 exp.inputString)
@@ -444,9 +455,10 @@ def decipher_primary_id(primaryId):
                     # Also get the ASN.1 type name as it is
                     # needed to build the Ada expression
                     # (Won't work for embedded types - FIXME)
-                    exp_typename = (exp.exprType.get('ReferencedTypeName') or
-                            exp.exprType['Kind']).replace('-', '_')
-                    if exp_type['Kind'] != 'ChoiceType':
+                    exp_typename = \
+                            (getattr(exp.exprType, 'ReferencedTypeName',
+                                 None) or exp.exprType.kind).replace('-', '_')
+                    if exp_type.kind != 'ChoiceType':
                         error = '{} is not a CHOICE'.format(exp.inputString)
                         LOG.error(error)
                         raise TypeError(error)
@@ -479,7 +491,6 @@ def decipher_expression(expr):
     local_decl = []
     if expr.kind == 'primary':
         if (expr.var.primType == None):
-               # or expr.var.primType.get('Kind') != 'ReferenceType'):
             # Populate the expression type to the field if it was not set
             # It can be the case if the type definition is embedded
             # in a SEQUENCE - in that case the ASN.1 compiler creates an
@@ -489,7 +500,7 @@ def decipher_expression(expr):
         stmts.extend(prim_stmts)
         local_decl.extend(local_var)
     elif expr.kind in ('plus', 'mul', 'minus', 'or', 'and', 'xor', 'eq',
-            'neq', 'gt', 'ge', 'lt', 'le', 'div', 'mod'):
+            'neq', 'gt', 'ge', 'lt', 'le', 'div', 'mod', 'rem'):
         left_stmts, left_str, left_local = decipher_expression(expr.left)
         right_stmts, right_str, right_local = decipher_expression(expr.right)
         ada_string = '({left} {op} {right})'.format(
@@ -507,11 +518,11 @@ def decipher_expression(expr):
            and expr.left.kind == 'primary'
            and expr.left.var.kind == 'stringLiteral'
            and len(expr.right.var.stringLiteral[1:-1]) +
-           len(expr.left.var.stringLiteral[1:-1]) > basic_type_expr['Max']):
+           len(expr.left.var.stringLiteral[1:-1]) > basic_type_expr.Max):
             raise ValueError('String concatenation exceeds container length: '
                     'length(' + expr.left.var.stringLiteral[1:-1] +
                     expr.right.var.stringLiteral[1:-1] + ') > ' +
-                    str(basic_type_expr['Max']))
+                    str(basic_type_expr.Max))
 
         left_stmts, left_str, left_local = decipher_expression(expr.left)
         right_stmts, right_str, right_local = decipher_expression(expr.right)
@@ -523,7 +534,7 @@ def decipher_expression(expr):
         ada_string = 'tmp{}'.format(expr.tmpVar)
         local_decl.append('{tmp} : aliased asn1Scc{eType};'.format(
                         tmp=ada_string,
-                        eType=expr.exprType['ReferencedTypeName']
+                        eType=expr.exprType.ReferencedTypeName
                         .replace('-', '_')))
 
         # If right or left is raw, declare a temporary variable for it, too
@@ -531,7 +542,7 @@ def decipher_expression(expr):
             if sexp.is_raw():
                 local_decl.append('tmp{idx} : aliased asn1Scc{eType};'.format(
                         idx=sexp.tmpVar,
-                        eType=sexp.exprType['ReferencedTypeName']
+                        eType=sexp.exprType.ReferencedTypeName
                         .replace('-', '_')))
                 stmts.append('tmp{idx} := {s_id};'.format(
                     idx=sexp.tmpVar, s_id=sid))
@@ -548,9 +559,9 @@ def decipher_expression(expr):
             else:
                 sexp.sid = sid
                 basic = find_basic_type(sexp.exprType)
-                if basic['Min'] == basic['Max']:
+                if basic.Min == basic.Max:
                     # Fixed-size string
-                    sexp.slen = basic['Max']
+                    sexp.slen = basic.Max
                 else:
                     # Variable-size types have a Length field
                     sexp.slen = '{}.Length'.format(sexp.sid)
@@ -575,7 +586,7 @@ def decipher_expression(expr):
         local_decl.extend(right_local)
         stmts.append("in_loop_{}:".format(ada_string))
         left_type = find_basic_type(expr.left.exprType)
-        if left_type['Min'] != left_type['Max']:
+        if left_type.Min != left_type.Max:
             stmts.append("for elem in 1..Integer({}.Length) loop"
                          .format(left_str))
         else:
@@ -587,10 +598,6 @@ def decipher_expression(expr):
         stmts.append("exit in_loop_{tmp} when {tmp} = True;"
                       .format(tmp=ada_string))
         stmts.append("end loop in_loop_{};".format(ada_string))
-
-    elif expr.kind == 'rem':
-        # TODO (Remove item from a set?)
-        pass
     else:
         raise ValueError('unsupported expression kind: ' + expr.kind)
     return stmts, ada_string, local_decl
@@ -604,13 +611,13 @@ def decipher_primary(primary):
     if primary.kind == 'primaryId':
         stmts, ada_string, local_decl = decipher_primary_id(primary.primaryId)
     elif primary.kind == 'enumeratedValue':
-        enumerant = primary.primaryId[0]
+        enumerant = primary.primaryId[0].replace('_', '-')
         basic = find_basic_type(primary.primType)
-        ada_string = ('asn1Scc' 
-                      + basic['EnumValues'][enumerant]['EnumID'])
+        ada_string = ('asn1Scc'
+                      + basic.EnumValues[enumerant].EnumID)
     elif primary.kind == 'choiceDeterminant':
-        enumerant = primary.primaryId[0]
-        ada_string = primary.primType['EnumValues'][enumerant]['EnumID']
+        enumerant = primary.primaryId[0].replace('_', '-')
+        ada_string = primary.primType.EnumValues[enumerant].EnumID
     elif primary.kind in ('numericalValue', 'booleanValue'):
         ada_string = primary.primaryId[0]
     elif primary.kind == 'sequence':
@@ -624,17 +631,17 @@ def decipher_primary(primary):
                 primary.choiceItem, primary.primType)
     elif primary.kind == 'emptyString':
         ada_string = 'asn1Scc{typeRef}_Init'.format(
-             typeRef=primary.primType['ReferencedTypeName'].replace('-', '_'))
+             typeRef=primary.primType.ReferencedTypeName.replace('-', '_'))
     elif primary.kind == 'stringLiteral':
         basic_type = find_basic_type(primary.primType)
-        if basic_type['Kind'].endswith('StringType'):
+        if basic_type.kind.endswith('StringType'):
             # If user put a literal string to fill an Octet string,
             # then convert the string to an array of unsigned_8 integers
             # as expected by the Ada type corresponding to Octet String
             unsigned_8 = [str(ord(val)) for val in primary.stringLiteral[1:-1]]
             ada_string = '(Data => (' + ', '.join(
                                                  unsigned_8) + ', others => 0)'
-            if basic_type['Min'] != basic_type['Max']:
+            if basic_type.Min != basic_type.Max:
                 # Non-fixed string size -> add Length field
                 ada_string += ', Length => {}'.format(
                                         str(len(primary.stringLiteral[1:-1])))
@@ -660,8 +667,7 @@ def decipher_if_then_else(ifThenElse, resType):
     # TODO: find a non-conflicing naming convention for tmp variable
     local_decl = ['tmp{idx} : asn1Scc{resType};'.format(
         idx=ifThenElse['tmpVar'],
-        resType=resType.get('ReferencedTypeName').replace('-', '_') or
-        'ERROR!')]
+        resType=resType.ReferencedTypeName.replace('-', '_'))]
     if_stmts, if_str, if_local = decipher_expression(ifThenElse['if'])
     then_stmts, then_str, then_local = decipher_expression(ifThenElse['then'])
     else_stmts, else_str, else_local = decipher_expression(ifThenElse['else'])
@@ -685,22 +691,23 @@ def decipher_if_then_else(ifThenElse, resType):
 
 def decipher_sequence(seq, seqType):
     ''' Return Ada string for an ASN.1 SEQUENCE '''
-    LOG.debug('Sequence: ' + str(seq) + str(seqType))
-    seq_ty = seqType.get('ReferencedTypeName', seqType)
+    LOG.debug('decipher_sequence: ' + str(seq) + str(seqType))
 
-    ada_string = 'asn1Scc{seqType}\'('.format(
-            seqType=seqType['ReferencedTypeName'].replace('-', '_'))
+    if seqType.kind == 'ReferenceType':
+        seq_ty = seqType.ReferencedTypeName
+    else:
+        seq_ty = seqType
+
+    ada_string = "asn1Scc{seqType}'(".format(
+            seqType=seqType.ReferencedTypeName.replace('-', '_'))
     stmts = []
     local_decl = []
     sep = ''
     for elem, value in seq.viewitems():
-        if value.exprType['Kind'] == 'UnknownType':
-            # Type may be unknown if it is an unnamed embedded type
-            # In that case, build up the type name by appending
-            # the last known type with the field name
-            value.exprType['Kind'] = 'ReferenceType'
-            value.exprType['ReferencedTypeName'] = '{thisType}_{elem}'.format(
-                    thisType=seqType['ReferencedTypeName'], elem=elem)
+        # Set the type of the field - easy thanks to ASN.1 flattened AST
+        delem = elem.replace('_', '-')
+        value.exprType = (TYPES
+                    [seqType.ReferencedTypeName].type.Children[delem].type)
         value_stmts, value_str, local_var = decipher_expression(value)
         ada_string += sep + elem + ' => ' + value_str
         sep = ', '
@@ -714,11 +721,11 @@ def decipher_sequence_of(seqof, seqofType):
     ''' Return Ada string for an ASN.1 SEQUENCE OF '''
     stmts = []
     local_decl = []
-    typename = seqofType.get('ReferencedTypeName')
+    typename = seqofType.ReferencedTypeName
     LOG.debug('SequenceOf Typename:' + str(typename))
-    asn_type = TYPES[typename]['type']
-    min_size = asn_type['Min']
-    max_size = asn_type['Max']
+    asn_type = TYPES[typename].type
+    min_size = asn_type.Min
+    max_size = asn_type.Max
     ada_string = 'asn1Scc{seqofType}\'('.format(
                                 seqofType=typename.replace('-', '_'))
     if min_size == max_size:
@@ -732,12 +739,8 @@ def decipher_sequence_of(seqof, seqofType):
                 .format(seqofType=typename.replace('-', '_'),
                         length=len(seqof)))
     for i in xrange(len(seqof)):
-        if seqof[i].primType is None:
-            # Embedded type (e.g. SEQUENCE OF SEQUENCE {...}):
-            # asn1Scc creates an intermediate type suffixed with _elm:
-            seqof[i].primType = {'Kind': 'ReferenceType',
-                    'ReferencedTypeName': '{baseType}_elm'.
-                    format(baseType=typename)}
+        # Set the type of the element
+        seqof[i].primType = TYPES[typename].type.type
         item_stmts, item_str, local_var = decipher_primary(seqof[i])
         stmts.extend(item_stmts)
         local_decl.extend(local_var)
@@ -748,15 +751,13 @@ def decipher_sequence_of(seqof, seqofType):
 
 def decipher_choice(choice, choiceType):
     ''' Return the Ada code for a CHOICE expression '''
-    if choice['value'].exprType['Kind'] == 'UnknownType':
-        # To handle non-explicit type definition (e.g CHOICE { a SEQUENCE...)
-        choice['value'].exprType = {'Kind': 'ReferenceType',
-                'ReferencedTypeName': '{baseType}_{choiceValue}'.format(
-                    baseType=choiceType.get('ReferencedTypeName'),
-                    choiceValue=choice['choice'])}
-
+    # Set the type of the choice field
+    # XXX TODO CHECK WHEN CHOICE TYPE IS BASIC
+    choice['value'].exprType = (
+          TYPES[choiceType.ReferencedTypeName].type.Children[choice['choice']])
     stmts, choice_str, local_decl = decipher_expression(choice['value'])
-    actual_type = choiceType.get('ReferencedTypeName') or choiceType['Kind']
+    actual_type = getattr(
+                     choiceType, 'ReferencedTypeName', None) or choiceType.kind
     actual_type = actual_type.replace('-', '_')
     ada_string = 'asn1Scc{cType}_{opt}_set({expr})'.format(
             cType=actual_type,
@@ -770,14 +771,16 @@ def decision_statement(dec):
     code = []
     local_decl = []
     question_type = dec.question.exprType
-    # Here is how to get properly the type (except when embedded in a sequence)
+    # Here is how to get properly the type
+    # (except when embedded in a sequence)
     # TODO fix the FIXMEs with that pattern:
-    actual_type = question_type.get(
-            'ReferencedTypeName') or question_type['Kind']
+    actual_type = getattr(
+            question_type, 'ReferencedTypeName', None) or question_type.kind
     actual_type = actual_type.replace('-', '_')
     basic = False
-    if actual_type in ('IntegerType', 'BooleanType',
-            'RealType', 'EnumeratedType'):
+    if actual_type in (
+                 'IntegerType', 'BooleanType', 'RealType', 'EnumeratedType',
+                 'ChoiceEnumeratedType'):
         basic = True
     # for ASN.1 types, declare a local variable
     # to hold the evaluation of the question
@@ -796,7 +799,10 @@ def decision_statement(dec):
         if a.kind == 'constant':
             a.kind = 'open_range'
             a.openRangeOp = 'eq'
-        if a.kind == 'open_range' and a.transition:
+        if a.kind == 'open_range':  # and a.transition:
+            # Note: removed and a.transition here because empty transitions
+            # have a different meaning, and a "null;" statement has to be
+            # generated, to go into the branch
             ans_stmts, ans_str, ans_decl = decipher_expression(a.constant)
             code.extend(ans_stmts)
             local_decl.extend(ans_decl)
@@ -813,18 +819,24 @@ def decision_statement(dec):
                 exp = '{q} {op} {ans}'.format(q=q_str,
                         op=OPERANDS[a.openRangeOp], ans=ans_str)
             code.append(sep + exp + ' then')
-            stmt, tr_decl = transition(a.transition)
+            if a.transition:
+                stmt, tr_decl = transition(a.transition)
+            else:
+                stmt, tr_decl = ['null;'], []
             code.extend(stmt)
             local_decl.extend(tr_decl)
             sep = 'elsif '
-        elif a.kind == 'close_range' and a.transition:
+        elif a.kind == 'close_range': #  and a.transition:
             sep = 'elsif '
             # TODO support close_range
         elif a.kind == 'informal_text':
             continue
-        elif a.kind == 'else' and a.transition:
+        elif a.kind == 'else':  # and a.transition:
             # Keep the ELSE statement for the end
-            else_code, else_decl = transition(a.transition)
+            if a.transition:
+                else_code, else_decl = transition(a.transition)
+            else:
+                else_code, else_decl = ['null;'], []
             local_decl.extend(else_decl)
     try:
         if sep != 'if ':
@@ -841,41 +853,100 @@ def decision_statement(dec):
     return code, local_decl
 
 
+def find_labels(trans):
+    '''
+        Yield a list of transition actions whenever a label is found
+        Used to transform labels into floating labels so that the gotos
+        in Ada can be contained within a single scope.
+    '''
+    if not trans:
+        return
+    # Terminators can have a label - add it to the transition actions
+    # (to trigger a goto at code generation)
+    if trans.terminator and trans.terminator.label:
+        trans.actions.append(trans.terminator.label)
+        trans.terminator.label = None
+    # Then for each action, check if there are labels and yield
+    # a new transition with the remaining actions (following the label)
+    for idx, action in enumerate(trans.actions):
+        if isinstance(action, ogAST.Label):
+            new_trans = ogAST.Transition()
+            # Create a floating label
+            flab = action
+            new_trans.actions = trans.actions[slice(idx+1, len(trans.actions))]
+            new_trans.terminator = trans.terminator
+            new_trans.terminators = trans.terminators
+            flab.transition = new_trans
+            # Transform the label into a JOIN in the original transition
+            trans.actions[idx:]=[]
+            trans.terminator = ogAST.Terminator()
+            trans.terminator.inputString = action.inputString
+            trans.terminator.kind = 'join'
+            # Recursively find labels in the new transition
+            for flabel in find_labels(flab.transition):
+                yield flabel
+            # Then yield the new transition
+            yield flab #new_trans
+        elif isinstance(action, ogAST.Decision):
+            for answer in action.answers:
+                for new_fl in find_labels(answer.transition):
+                    # Append the remaining actions of the transition
+                    if not new_fl.transition.terminator:
+                        new_fl.transition.actions.extend(
+                               trans.actions[slice(idx+1, len(trans.actions))])
+                        new_fl.transition.terminator = trans.terminator
+                    yield new_fl
+
+
 def transition(tr):
     ''' generate the code for a transition '''
-    code = []
-    local_decl = []
+    code, local_decl = [], []
+    empty_transition = True
     for action in tr.actions:
         if isinstance(action, ogAST.Output):
             stmt, local_var = output_statement(action)
+            code.extend(stmt)
+            local_decl.extend(local_var)
+            empty_transition = False
         elif isinstance(action, ogAST.Task):
             stmt, local_var = task_statement(action)
+            code.extend(stmt)
+            local_decl.extend(local_var)
+            if not action.informalText:
+                empty_transition = False
         elif isinstance(action, ogAST.Decision):
             stmt, local_var = decision_statement(action)
+            code.extend(stmt)
+            local_decl.extend(local_var)
+            empty_transition = False
         elif isinstance(action, ogAST.Label):
-            stmt = ['<<{label}>>'.format(
-                label=action.inputString)]
-            local_var = []
-        code.extend(stmt)
-        local_decl.extend(local_var)
-    if tr.terminator:
-        code.extend(traceability(tr.terminator))
-        if tr.terminator.label:
-            code.append('<<{label}>>'.format(
-                label=tr.terminator.label.inputString))
-        if tr.terminator.kind == 'next_state':
-            if tr.terminator.inputString.strip() != '-':
-                # discard the dash state (remain in the same state)
-                code.append('state := {nextState};'.format(
-                             nextState=tr.terminator.inputString))
-            # In any case, return to avoid code of floating labels
-            code.append('return;')
-        elif tr.terminator.kind == 'join':
-            code.append('goto {label};'.format(
-                label=tr.terminator.inputString))
-        elif tr.terminator.kind == 'stop':
-            pass
-            # TODO
+            # Transition following labels are generated in a separate section
+            # for visibility reasons (see Ada scope)
+            code.append('goto {label};'.format(label=action.inputString))
+            break
+    else:
+        if tr.terminator:
+            empty_transition = False
+            code.extend(traceability(tr.terminator))
+            if tr.terminator.label:
+                code.append('<<{label}>>'.format(
+                    label=tr.terminator.label.inputString))
+            if tr.terminator.kind == 'next_state':
+                if tr.terminator.inputString.strip() != '-':
+                    # discard the dash state (remain in the same state)
+                    code.append('state := {nextState};'.format(
+                                 nextState=tr.terminator.inputString))
+                # In any case, return to avoid code of floating labels
+                code.append('return;')
+            elif tr.terminator.kind == 'join':
+                code.append('goto {label};'.format(
+                    label=tr.terminator.inputString))
+            elif tr.terminator.kind == 'stop':
+                pass
+                # TODO
+    if empty_transition:
+        # If transition does not have any statement, generate an Ada 'null;'
+        code.append('null;')
     return code, local_decl
 
 
@@ -901,16 +972,24 @@ def inner_procedure(proc):
     # with procedure defined inside the current procedure
     # Not critical: the editor forbids procedures inside procedures
 
+    # Save variable scope (as local variables may shadow process variables)
+    outer_scope = dict(VARIABLES)
+    VARIABLES.update(proc.variables)
+    # Also add procedure parameters in scope
+    for var in proc.fpar:
+        VARIABLES.update({var['name']: var['type']})
+
     # Build the procedure signature
     pi_header = 'procedure {proc_name}'.format(proc_name=proc.inputString)
     if proc.fpar:
         pi_header += '('
         params = []
         for fpar in proc.fpar:
+            typename = fpar['type'].ReferencedTypeName.replace('-', '_')
             params.append('l_{name}: in{out} asn1Scc{ptype}'.format(
                     name=fpar.get('name'),
                     out=' out' if fpar.get('direction') == 'out' else '',
-                    ptype=fpar.get('type')))
+                    ptype=typename))
         pi_header += ';'.join(params)
         pi_header += ')'
 
@@ -926,8 +1005,9 @@ def inner_procedure(proc):
             code.extend(inner_code)
         code.append(pi_header + ' is')
         for var_name, var_type in proc.variables.viewitems():
+            typename = var_type.ReferencedTypeName.replace('-', '_')
             code.append('l_{name} : asn1Scc{sort};'.format(
-                name=var_name, sort=var_type))
+                name=var_name, sort=typename))
         tr_code, tr_decl = transition(proc.content.start.transition)
         # Generate code for the floating labels
         code_labels = []
@@ -942,8 +1022,11 @@ def inner_procedure(proc):
         code.append('end {procName};'.format(procName=proc.inputString))
     code.append('\n')
 
-    return code, local_decl
+    # Reset the scope to how it was prior to the procedure definition
+    VARIABLES.clear()
+    VARIABLES.update(outer_scope)
 
+    return code, local_decl
 
 def format_ada_code(stmts):
     ''' Indent properly the Ada code '''
@@ -953,14 +1036,14 @@ def format_ada_code(stmts):
         elems = line.strip().split()
         if elems and elems[0] in ('when', 'end', 'elsif', 'else'):
             indent = max(indent - 1, 0)
-        if elems and elems[-1] == 'case;':
+        if elems and elems[-1] == 'case;': # Corresponds to end case;
             indent = max(indent - 1, 0)
         yield indent_pattern * indent + line
         if elems and elems[-1] in ('is', 'then', 'loop'):
             indent += 1
         if elems and elems[0] in ('begin', 'case', 'else', 'when'):
             indent += 1
-        if not elems:
+        if not elems:  # newline -> decrease indent
             indent -= 1
     yield stmts[-1]
 
@@ -984,7 +1067,7 @@ def generate(process):
     process_level_decl = []
     for var_name, var_type in VARIABLES.iteritems():
         process_level_decl.append('l_{n} : aliased asn1Scc{t};'.format(
-            n=var_name, t=var_type.replace('-','_')))
+            n=var_name, t=var_type.ReferencedTypeName.replace('-','_')))
 
     # Add the process states list to the process-level variables
     states_decl = 'type states is ('
@@ -992,6 +1075,11 @@ def generate(process):
     process_level_decl.append(states_decl)
     process_level_decl.append('state : states := START;')
 
+    # Add function allowing to trace current state as a string
+    process_level_decl.append('function get_state return String;')
+    process_level_decl.append('pragma export(C, get_state, "{}_state");'
+                                                         .format(process_name))
+
     # Add the declaration of the runTransition procedure
     process_level_decl.append('procedure runTransition(trId: Integer);')
 
@@ -1061,18 +1149,19 @@ package {process_name} is'''.format(process_name=process_name,
 
     # Generate the code for each input signal (provided interface) and timers
     for signal in process.input_signals + [
-                        {'name': timer.lower() for timer in process.timers}]:
+                        {'name': timer.lower()} for timer in process.timers]:
         if signal.get('name', 'START') == 'START':
             continue
         pi_header = 'procedure {sig_name}'.format(sig_name=signal['name'])
         param_name = signal.get('param_name') or 'MISSING_PARAM_NAME'
         # Add (optional) PI parameter (only one is possible in TASTE PI)
         if 'type' in signal:
+            typename = signal['type'].ReferencedTypeName.replace('-', '_')
             pi_header += '({pName}: access asn1Scc{pType})'.format(
-                pName=param_name, pType=signal['type'])
+                                           pName=param_name, pType=typename)
 
         # Add declaration of the provided interface in the .ads file
-        ads_template.append('-- Provided interface "' + signal['name'] + '"')
+        ads_template.append('--  Provided interface "' + signal['name'] + '"')
         ads_template.append(pi_header + ';')
 
         pi_header += ' is'
@@ -1094,7 +1183,6 @@ package {process_name} is'''.format(process_name=process_name,
                 if input_def.transition:
                     taste_template.append('runTransition({idx});'.format(
                         idx=input_def.transition_id))
-                        #t=input_def.inputlist[signal['name']]))
                 else:
                     taste_template.append('null;')
             else:
@@ -1113,10 +1201,11 @@ package {process_name} is'''.format(process_name=process_name,
         param_name = signal.get('param_name') or 'MISSING_PARAM_NAME'
         # Add (optional) RI parameter
         if 'type' in signal:
-            ri_header += '({pName}: access asn1Scc{pType});'.format(
-                pName=param_name, pType=signal['type'])
-        ads_template.append('-- Required interface "' + signal['name'] + '"')
-        ads_template.append(ri_header)
+            typename = signal['type'].ReferencedTypeName.replace('-', '_')
+            ri_header += '({pName}: access asn1Scc{pType})'.format(
+                pName=param_name, pType=typename)
+        ads_template.append('--  Required interface "' + signal['name'] + '"')
+        ads_template.append(ri_header + ';')
         ads_template.append('pragma import(C, {sig}, "{proc}_RI_{sig}");'
                 .format(sig=signal['name'], proc=process_name))
 
@@ -1125,21 +1214,47 @@ package {process_name} is'''.format(process_name=process_name,
         ri_header = 'procedure {sig_name}'.format(sig_name=proc.inputString)
         params = []
         for param in proc.fpar:
-            params.append('{par[name]}: access asn1Scc{par[type]}'.format(
-                par=param))
+            typename = param['type'].ReferencedTypeName.replace('-', '_')
+            params.append('{par[name]}: access asn1Scc{partype}'.format(
+                par=param, partype=typename))
         if params:
-            ri_header += '(' + ';'.join(params) + ');'
-        ads_template.append('-- Sync required interface "' + proc.inputString)
-        ads_template.append(ri_header)
+            ri_header += '(' + ';'.join(params) + ')'
+        ads_template.append('--  Sync required interface "' + proc.inputString)
+        ads_template.append(ri_header + ';')
         ads_template.append('pragma import(C, {sig}, "{proc}_RI_{sig}");'
                 .format(sig=proc.inputString, proc=process_name))
 
+    # for the .ads file, generate the declaration of timers set/reset functions
+    for timer in process.timers:
+        ads_template.append(
+                '--  Timer {} SET and RESET functions'.format(timer))
+        ads_template.append('procedure SET_{}(val: access asn1SccT_UInt32);'
+                .format(timer))
+        ads_template.append(
+                'pragma import(C, SET_{timer}, "{proc}_RI_set_{timer}");'
+                .format(timer=timer, proc=process_name))
+        ads_template.append('procedure RESET_{};'.format(timer))
+        ads_template.append(
+                'pragma import(C, RESET_{timer}, "{proc}_RI_reset_{timer}");'
+                .format(timer=timer, proc=process_name))
+
+
     taste_template.append('procedure runTransition(trId: Integer) is')
 
     # Generate the code for all transitions
     code_transitions = []
     local_decl_transitions = []
+
+    # Transform inner labels of both floating labels and transitions
+    # into new floating labels, so that they get generated in a separate
+    # section of the Ada code, where they are in the scope of everybody
+    for idx in xrange(len(process.content.floating_labels)):
+        for new_floating in find_labels(
+                              process.content.floating_labels[idx].transition):
+            process.content.floating_labels.append(new_floating)
     for proc_tr in process.transitions:
+        for new_floating in find_labels(proc_tr):
+            process.content.floating_labels.append(new_floating)
         code_tr, tr_local_decl = transition(proc_tr)
         code_transitions.append(code_tr)
         local_decl_transitions.extend(tr_local_decl)
@@ -1179,6 +1294,14 @@ package {process_name} is'''.format(process_name=process_name,
     taste_template.append('end runTransition;')
     taste_template.append('\n')
 
+    # Code of the function allowing to trace current state
+    taste_template.append('function get_state return String is')
+    taste_template.append('begin')
+    taste_template.append("return states'Image(state);")
+    taste_template.append('end get_state;')
+    taste_template.append('\n')
+
+
     taste_template.extend(start_transition)
     taste_template.append('end {process_name};'
             .format(process_name=process_name))
