@@ -6,7 +6,7 @@
 
     SDL is the Specification and Description Language (Z100 standard from ITU)
 
-    Copyright (c) 2012 European Space Agency
+    Copyright (c) 2012-2013 European Space Agency
 
     Designed and implemented by Maxime Perrotin
 
@@ -121,10 +121,10 @@ def paste_floating_objects(scene):
         for state in states:
             # First check if state has already been pasted
             try:
-                new_item = Renderer.render_state(
-                        CLIPBOARD, state, terminators, states)
+                new_item = Renderer.render(state, scene=CLIPBOARD,
+                           terminators=terminators, states=states)
             except TypeError:
-                # Discard terminators (explanation given in render_state)
+                # Discard terminators (explanation given in Renderer._state)
                 pass
             else:
                 LOG.debug('PASTE STATE "' + state.inputString + '"')
@@ -133,7 +133,7 @@ def paste_floating_objects(scene):
                 scene.addItem(new_item)
         for text_area in text_areas:
             LOG.debug('PASTE TEXT AREA')
-            new_item = Renderer.render_text_area(scene, text_area)
+            new_item = Renderer.render(text_area, scene)
             symbols.add(new_item)
     return symbols
 
@@ -148,15 +148,8 @@ def paste_below_item(parent, scene):
                 (c, (ogAST.State, ogAST.TextArea, ogAST.Start))]:
             LOG.debug('PASTE ' + str(i))
             # Create the new item from the AST description
-            if isinstance(i, ogAST.Input):
-                # use absolute coordinates
-                new_item = Renderer.render_input(
-                        CLIPBOARD,
-                        i, parent=None, states=states)
-            else:
-                new_item = Renderer.render_action(
-                        CLIPBOARD,
-                        i, parent=None, states=states)
+            new_item = Renderer.render(i, scene=CLIPBOARD,
+                                       parent=None, states=states)
             # Check that item is compatible with parent
             if (type(new_item).__name__ in parent.allowed_followers):
                 # Move the item from the clipboard to the scene
