@@ -70,7 +70,7 @@ class Expression(object):
                 and self.var.kind not in ('expression', 'constant')
                 and (self.var.kind != 'primaryId' or
                     (len(self.var.primaryId) < 2
-                     and self.exprType['Kind'] != 'ReferenceType')))
+                     and self.exprType.kind != 'ReferenceType')))
 
 
 class Primary(object):
@@ -327,6 +327,14 @@ class Transition(object):
         self.terminators = []
 
 
+    def __repr__(self):
+        ''' Debug output: display all actions '''
+        data = [repr(action) for action in self.actions]
+        if self.terminator:
+            data.append(repr(self.terminator))
+        return '\n'.join(data)
+
+
 class Input(object):
     ''' AST Entry for the INPUT symbol '''
     def __init__(self):
@@ -367,6 +375,7 @@ class Start(object):
     ''' AST Entry for the START symbol '''
     def __init__(self):
         ''' Initialize the Start symbol attributes '''
+        self.inputString = ''
         self.pos_x = 0
         self.pos_y = 0
         self.width = 100
@@ -443,7 +452,8 @@ class TextArea(object):
     ''' AST Entry for text areas (containing declarations/comments) '''
     def __init__(self):
         ''' Text area (raw content for rendering only) '''
-        self.inputString = '-- Declare your variables\n\nDCL foo MyInteger;'
+        self.inputString = '-- Declare your variables\n\n' \
+                           '-- Syntax: DCL <variable name> <type name>;'
         # DCL variables in the text area {name: sort, ...}
         self.variables = {}
         self.line = None
@@ -502,7 +512,7 @@ class Procedure(object):
         # Keep a list of labels and floating labels - useful for backends
         self.labels = []
         #self.floating_labels = []
-        # Inherited procedures and operators
+        # Inherited procedures and operators XXX should be lists no?
         self.procedures = {}
         self.operators = {}
         # inner procedures and operators (see Process for format)
@@ -540,7 +550,6 @@ class Process(object):
         self.asn1Modules = None
         self.dataview = None
 
-        # input and output signal lists: {'signal': definition from iv.py, ...}
         # input and output signal lists:
         # [{'name': str, 'type': str, 'direction':'in'/'out'}]
         self.input_signals = []
@@ -568,6 +577,7 @@ class Process(object):
         self.transitions = []
 
         # Set of symbols contained in the process (type Automaton)
+        # (Includes inner procedures)
         self.content = Automaton(parent=self)
 
         # List of timers (strings) declared in the process
